@@ -69,9 +69,14 @@ static struct {
 	char *device_udid;
 #ifdef HAVE_LIBIMOBILEDEVICE_1_1
 	char *appid;
+	int use_container;
 #endif
 	char *service_name;
+#ifdef HAVE_LIBIMOBILEDEVICE_1_1_5
+	lockdownd_service_descriptor_t service;
+#else
 	uint16_t port;
+#endif
 } opts;
 
 enum {
@@ -80,9 +85,10 @@ enum {
 	KEY_ROOT,
 	KEY_UDID,
 	KEY_UDID_LONG,
-	KEY_APPID,
-	KEY_APPID_LONG,
-	KEY_DEBUG
+	KEY_VENDOR_DOCUMENTS_LONG,
+	KEY_VENDOR_CONTAINER_LONG,
+	KEY_DEBUG,
+	KEY_DEBUG_LONG
 };
 
 static struct fuse_opt ifuse_opts[] = {
@@ -93,10 +99,11 @@ static struct fuse_opt ifuse_opts[] = {
 	FUSE_OPT_KEY("-u %s",          KEY_UDID),
 	FUSE_OPT_KEY("--udid %s",      KEY_UDID_LONG),
 	FUSE_OPT_KEY("--root",         KEY_ROOT),
-	FUSE_OPT_KEY("--debug",        KEY_DEBUG),
+	FUSE_OPT_KEY("-d",             KEY_DEBUG),
+	FUSE_OPT_KEY("--debug",        KEY_DEBUG_LONG),
 #ifdef HAVE_LIBIMOBILEDEVICE_1_1
-	FUSE_OPT_KEY("-a %s",          KEY_APPID),
-	FUSE_OPT_KEY("--appid %s",     KEY_APPID_LONG),
+	FUSE_OPT_KEY("--documents %s", KEY_VENDOR_DOCUMENTS_LONG),
+	FUSE_OPT_KEY("--container %s", KEY_VENDOR_CONTAINER_LONG),
 #endif
 	FUSE_OPT_END
 };
@@ -118,30 +125,30 @@ struct afc_error_mapping {
 	afc_error_t from;
 	int to;
 } static afc_error_to_errno_map[] = {
-	{AFC_E_SUCCESS			, 0},
-	{AFC_E_OP_HEADER_INVALID	, EIO},
-	{AFC_E_NO_RESOURCES		, EMFILE},
-	{AFC_E_READ_ERROR		, ENOTDIR},
-	{AFC_E_WRITE_ERROR		, EIO},
-	{AFC_E_UNKNOWN_PACKET_TYPE	, EIO},
-	{AFC_E_INVALID_ARG		, EINVAL},
-	{AFC_E_OBJECT_NOT_FOUND		, ENOENT},
-	{AFC_E_OBJECT_IS_DIR		, EISDIR},
-	{AFC_E_DIR_NOT_EMPTY		, ENOTEMPTY},
-	{AFC_E_PERM_DENIED		, EPERM},
-	{AFC_E_SERVICE_NOT_CONNECTED	, ENXIO},
-	{AFC_E_OP_TIMEOUT		, ETIMEDOUT},
-	{AFC_E_TOO_MUCH_DATA		, EFBIG},
-	{AFC_E_END_OF_DATA		, ENODATA},
-	{AFC_E_OP_NOT_SUPPORTED		, ENOSYS},
-	{AFC_E_OBJECT_EXISTS		, EEXIST},
-	{AFC_E_OBJECT_BUSY		, EBUSY},
-	{AFC_E_NO_SPACE_LEFT		, ENOSPC},
-	{AFC_E_OP_WOULD_BLOCK		, EWOULDBLOCK},
-	{AFC_E_IO_ERROR			, EIO},
-	{AFC_E_OP_INTERRUPTED		, EINTR},
-	{AFC_E_OP_IN_PROGRESS		, EALREADY},
-	{AFC_E_INTERNAL_ERROR		, EIO},
+	{AFC_E_SUCCESS               , 0},
+	{AFC_E_OP_HEADER_INVALID     , EIO},
+	{AFC_E_NO_RESOURCES          , EMFILE},
+	{AFC_E_READ_ERROR            , ENOTDIR},
+	{AFC_E_WRITE_ERROR           , EIO},
+	{AFC_E_UNKNOWN_PACKET_TYPE   , EIO},
+	{AFC_E_INVALID_ARG           , EINVAL},
+	{AFC_E_OBJECT_NOT_FOUND      , ENOENT},
+	{AFC_E_OBJECT_IS_DIR         , EISDIR},
+	{AFC_E_DIR_NOT_EMPTY         , ENOTEMPTY},
+	{AFC_E_PERM_DENIED           , EPERM},
+	{AFC_E_SERVICE_NOT_CONNECTED , ENXIO},
+	{AFC_E_OP_TIMEOUT            , ETIMEDOUT},
+	{AFC_E_TOO_MUCH_DATA         , EFBIG},
+	{AFC_E_END_OF_DATA           , ENODATA},
+	{AFC_E_OP_NOT_SUPPORTED      , ENOSYS},
+	{AFC_E_OBJECT_EXISTS         , EEXIST},
+	{AFC_E_OBJECT_BUSY           , EBUSY},
+	{AFC_E_NO_SPACE_LEFT         , ENOSPC},
+	{AFC_E_OP_WOULD_BLOCK        , EWOULDBLOCK},
+	{AFC_E_IO_ERROR              , EIO},
+	{AFC_E_OP_INTERRUPTED        , EINTR},
+	{AFC_E_OP_IN_PROGRESS        , EALREADY},
+	{AFC_E_INTERNAL_ERROR        , EIO},
 	{-1}
 };
 
@@ -246,6 +253,11 @@ static int ifuse_getattr(const char *path, struct stat *stbuf)
 			} else if (!strcmp(info[i], "st_mtime")) {
 				stbuf->st_mtime = (time_t)(atoll(info[i+1]) / 1000000000);
 			}
+#ifdef _DARWIN_FEATURE_64_BIT_INODE
+			else if (!strcmp(info[i], "st_birthtime")) { /* available on iOS 7+ */
+				stbuf->st_birthtime = (time_t)(atoll(info[i+1]) / 1000000000);
+			}
+#endif
 		}
 		free_dictionary(info);
 
@@ -402,7 +414,11 @@ void *ifuse_init(struct fuse_conn_info *conn)
 		afc_client_new_from_house_arrest_client(house_arrest, &afc);
 	} else { 
 #endif
+#ifdef HAVE_LIBIMOBILEDEVICE_1_1_5
+		afc_client_new(phone, opts.service, &afc);
+#else
 		afc_client_new(phone, opts.port, &afc);
+#endif
 #ifdef HAVE_LIBIMOBILEDEVICE_1_1
 	}
 #endif
@@ -611,17 +627,18 @@ static struct fuse_operations ifuse_oper = {
 
 static void print_usage()
 {
-	fprintf(stderr, "Usage: ifuse <mount_point> [OPTIONS]\n");
-	fprintf(stderr, "Mount filesystem of an iPhone, iPod Touch, iPad or Apple TV.\n\n");
+	fprintf(stderr, "Usage: " PACKAGE_NAME " MOUNTPOINT [OPTIONS]\n");
+	fprintf(stderr, "Mount directories of an iOS device locally using fuse.\n\n");
 	fprintf(stderr, "  -o opt,[opt...]\tmount options\n");
 	fprintf(stderr, "  -u, --udid UDID\tmount specific device by its 40-digit device UDID\n");
 	fprintf(stderr, "  -h, --help\t\tprint usage information\n");
 	fprintf(stderr, "  -V, --version\t\tprint version\n");
+	fprintf(stderr, "  -d, --debug\t\tenable libimobiledevice communication debugging\n");
 #ifdef HAVE_LIBIMOBILEDEVICE_1_1
-	fprintf(stderr, "  --appid APPID\t\tmount 'Documents' folder of app identified by APPID\n");
+	fprintf(stderr, "  --documents APPID\tmount 'Documents' folder of app identified by APPID\n");
+	fprintf(stderr, "  --container APPID\tmount sandbox root of an app identified by APPID\n");
 #endif
 	fprintf(stderr, "  --root\t\tmount root file system (jailbroken device required)\n");
-	fprintf(stderr, "  --debug\t\tenable libimobiledevice communication debugging\n");
 	fprintf(stderr, "\n");
 	fprintf(stderr, "Example:\n\n");
 	fprintf(stderr, "  $ ifuse /media/iPhone --root\n\n");
@@ -645,18 +662,20 @@ static int ifuse_opt_proc(void *data, const char *arg, int key, struct fuse_args
 		res = 0;
 		break;
 #ifdef HAVE_LIBIMOBILEDEVICE_1_1
-	case KEY_APPID_LONG:
-		opts.appid = strdup(arg+7);
+	case KEY_VENDOR_CONTAINER_LONG:
+		opts.use_container = 1;
+		opts.appid = strdup(arg+11);
 		opts.service_name = HOUSE_ARREST_SERVICE_NAME;
 		res = 0;
 		break;
-	case KEY_APPID:
-		opts.appid = strdup(arg+2);
+	case KEY_VENDOR_DOCUMENTS_LONG:
+		opts.appid = strdup(arg+11);
 		opts.service_name = HOUSE_ARREST_SERVICE_NAME;
 		res = 0;
 		break;
 #endif
 	case KEY_DEBUG:
+	case KEY_DEBUG_LONG:
 		idevice_set_debug_level(1);
 		res = 0;
 		break;
@@ -668,7 +687,7 @@ static int ifuse_opt_proc(void *data, const char *arg, int key, struct fuse_args
 		print_usage();
 		exit(EXIT_SUCCESS);
 	case KEY_VERSION:
-		fprintf(stderr, "%s version %s\n", PACKAGE_NAME, PACKAGE_VERSION);
+		fprintf(stderr, "%s %s\n", PACKAGE_NAME, PACKAGE_VERSION);
 		exit(EXIT_SUCCESS);
 	case FUSE_OPT_KEY_OPT:
 		/* ignore other options and pass them to fuse_main later */
@@ -737,6 +756,11 @@ int main(int argc, char *argv[])
 			fprintf(stderr, "Please disable the password protection on your device and try again.\n");
 			fprintf(stderr, "The device does not allow pairing as long as a password has been set.\n");
 			fprintf(stderr, "You can enable it again after the connection succeeded.\n");
+#ifdef LOCKDOWN_E_PAIRING_DIALOG_PENDING
+		} else if (ret == LOCKDOWN_E_PAIRING_DIALOG_PENDING) {
+			fprintf(stderr, "Please dismiss the trust dialog on your device and try again.\n");
+			fprintf(stderr, "The device does not allow pairing as long as the dialog has not been accepted.\n");
+#endif
 		} else {
 			fprintf(stderr, "Failed to connect to lockdownd service on the device.\n");
 			fprintf(stderr, "Try again. If it still fails try rebooting your device.\n");
@@ -744,7 +768,13 @@ int main(int argc, char *argv[])
 		return EXIT_FAILURE;
 	}
 
-	if ((lockdownd_start_service(control, opts.service_name, &opts.port) != LOCKDOWN_E_SUCCESS) || !opts.port) {
+	if (
+#ifdef HAVE_LIBIMOBILEDEVICE_1_1_5
+	(lockdownd_start_service(control, opts.service_name, &opts.service) != LOCKDOWN_E_SUCCESS) || !opts.service
+#else
+	(lockdownd_start_service(control, opts.service_name, &opts.port) != LOCKDOWN_E_SUCCESS) || !opts.port
+#endif
+ 	) {
 		lockdownd_client_free(control);
 		idevice_free(phone);
 		fprintf(stderr, "Failed to start AFC service '%s' on the device.\n", opts.service_name);
@@ -757,13 +787,19 @@ int main(int argc, char *argv[])
 
 #ifdef HAVE_LIBIMOBILEDEVICE_1_1
 	if (!strcmp(opts.service_name, HOUSE_ARREST_SERVICE_NAME)) {
+#ifdef HAVE_LIBIMOBILEDEVICE_1_1_5
+		house_arrest_client_new(phone, opts.service, &house_arrest);
+#else
 		house_arrest_client_new(phone, opts.port, &house_arrest);
+#endif
 		if (!house_arrest) {
 			fprintf(stderr, "Could not start document sharing service!\n");
 			return EXIT_FAILURE;
 		}
-		if (house_arrest_send_command(house_arrest, "VendContainer", opts.appid) != HOUSE_ARREST_E_SUCCESS) {
-			fprintf(stderr, "Could not send VendContainer command!\n");
+
+		/* FIXME: iOS 3.x house_arrest does not know about VendDocuments yet, thus use VendContainer and chroot manually with fuse subdir module */
+		if (house_arrest_send_command(house_arrest, opts.use_container ? "VendContainer": "VendDocuments", opts.appid) != HOUSE_ARREST_E_SUCCESS) {
+			fprintf(stderr, "Could not send house_arrest command!\n");
 			goto leave_err;
 		}
 
@@ -782,8 +818,10 @@ int main(int argc, char *argv[])
 		}
 		plist_free(dict);
 
-		fuse_opt_add_arg(&args, "-omodules=subdir");
-		fuse_opt_add_arg(&args, "-osubdir=Documents");
+		if (opts.use_container == 0) {
+			fuse_opt_add_arg(&args, "-omodules=subdir");
+			fuse_opt_add_arg(&args, "-osubdir=Documents");
+		}
 	}
 #endif
 	res = fuse_main(args.argc, args.argv, &ifuse_oper, NULL);
